/*
 * flawd.c
 *
 * Copyright Stephen C. Tweedie, 1997 
 *
 * Redistribution of this file is permitted under the GNU Public
 * License.
 *
 * A simple ramdisk driver module for debugging purposes.  It allows 
 * blocks to be forced bad at will, and also permits random errors to be
 * injected into IO requests.
 */

/* For now, this device is always a module */
#define MODULE

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#include <asm/segment.h>
#include <asm/bitops.h>

#include "flawd.h"

#define MAJOR_NR FLAWD_MAJOR

#define DEVICE_NAME "flawd"
#define DEVICE_REQUEST do_flawd_request
#define DEVICE_NR(device) (MINOR(device))
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#define TIMEOUT_VALUE (6 * HZ)

#include <linux/blk.h>

static struct flawd_device flawd_minor_table[NR_FLAWD_MINOR];
static int flawd_sizes[NR_FLAWD_MINOR];
static int flawd_blksizes[NR_FLAWD_MINOR];


/*
 * The strategy function for the flawd device driver.  We handle all new
 * requests here.
 */
static void do_flawd_request(void)
{
	int	current_sector, start, len;
	char	*user_addr, *flawd_addr;

	/*	
	int	real_block, block, offset, len, blksize, size;
	 */
	kdev_t	devnr;
	struct flawd_device *flawd;
	struct buffer_head *bh;

repeat:
	/*
	 * Fetch the next request of the request queue
	 */
	INIT_REQUEST;
	if (MINOR(CURRENT->rq_dev) >= NR_FLAWD_MINOR)
		goto error_out;

	/*
	 * Locate our flawd device
	 */
	flawd = &flawd_minor_table[MINOR(CURRENT->rq_dev)];

	/*
	 * Check that the request fits into the flawd device!
	 */
	current_sector = CURRENT->sector;
	len	       = CURRENT->current_nr_sectors;

#if 0
	printk ("flawd: %s %d plus %d\n",
		(CURRENT->cmd == READ) ? "read" : "write",
		current_sector, len);
#endif

	if (current_sector + len > flawd->nr_sectors)
		goto error_out;

	/*
	 * Convert from sector numbers to byte offsets.
	 */
	start = current_sector << 9;
	len   = len << 9;

	/*
	 * Locate the transfer addresses
	 */
	flawd_addr  = flawd->data + start;
	user_addr = CURRENT->buffer;
	
	while (len > 0) {
		/*
		 * First of all check whether we need to induce an IO error
		 * here 
		 */
		if (test_bit (current_sector, (int *) flawd->bad_block_map))
			goto error_out;

		switch (CURRENT->cmd) {
		case READ:
			memcpy (user_addr, flawd_addr, 512);
			break;
		case WRITE:
			memcpy (flawd_addr, user_addr, 512);
			break;
		default:
			printk ("flawd: Unrecognised command %d\n", 
				CURRENT->cmd);
			goto error_out;
		}
				
		user_addr  += 512;
		flawd_addr += 512;
		len        -= 512;
		
		current_sector++;
	}
	end_request(1);
	goto repeat;
error_out:
	end_request(0);
	goto repeat;
}


static int get_flawd_device (struct inode * inode, struct flawd_device **p)
{
	struct flawd_device *flawd;
	int minor;
	
	if (!inode)
		return -EINVAL;
	if (MAJOR(inode->i_rdev) != MAJOR_NR) {
		printk("get_flawd_device: pseudo-major != %d\n", MAJOR_NR);
		return -ENODEV;
	}
	minor = MINOR(inode->i_rdev);
	if (minor >= NR_FLAWD_MINOR)
		return -ENODEV;
	flawd = &flawd_minor_table[minor];
	if (!flawd->nr_sectors)
		return -ENODEV;
	
	*p = flawd;
	return 0;
}

static int flawd_ioctl(struct inode * inode, struct file * file,
		       unsigned int cmd, unsigned long arg)
{
	struct flawd_device *flawd;
	int err;
	int start, count;
	__FLAWD_SECTOR_RANGE *farg;
	int i;
	
	err = get_flawd_device(inode, &flawd);
	if (err)
		return err;

	switch (cmd) {
	case FLAWD_SET_BAD_SECTORS :
	case FLAWD_CLEAR_BAD_SECTORS :
		farg = (__FLAWD_SECTOR_RANGE *) arg;
#if LINUX_VERSION_CODE >= LinuxVersionCode(2,1,0)
		err = get_user(start, &farg->start) ||
			get_user(count, &farg->count);
		if (err)
			return err;
#else
		err = verify_area (VERIFY_READ, 
				   (void *) arg,
				   sizeof (__FLAWD_SECTOR_RANGE));
		if (err)
			return err;
		start = get_user (&farg->start);
		count = get_user (&farg->count);
#endif
		if (start < 0 || count < 0 ||
		    start + count > flawd->nr_sectors)
			return -EINVAL;
		
		if (cmd == FLAWD_SET_BAD_SECTORS) {
			for (i = count; i > 0; i--) 
				set_bit (start++, flawd->bad_block_map);
		} else {
			for (i = count; i > 0; i--) 
				clear_bit (start++, flawd->bad_block_map);
		}
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int flawd_open(struct inode *inode, struct file *file)
{
	struct flawd_device *flawd;
	int err;

	err = get_flawd_device(inode, &flawd);
	if (err)
		return err;
	flawd->refcnt++;
	MOD_INC_USE_COUNT;
	return 0;
}


static RELEASE_RET_TYPE flawd_release(struct inode *inode, struct file *file)
{
	struct flawd_device *flawd;
	int err;

	err = get_flawd_device(inode, &flawd);
	if (err)
		RELEASE_RETURN;
	fsync_dev(inode->i_rdev);
	if (flawd->refcnt <= 0)
		printk("flawd_release: refcount(%d) <= 0\n", flawd->refcnt);
	else  {
		flawd->refcnt--;
		MOD_DEC_USE_COUNT;
	}
	RELEASE_RETURN;
}


static struct file_operations flawd_fops = {
	NULL,			/* lseek - default */
	block_read,		/* read - general block-dev read */
	block_write,		/* write - general block-dev write */
	NULL,			/* readdir - bad */
	NULL,			/* poll */
	flawd_ioctl,		/* ioctl */
	NULL,			/* mmap */
	flawd_open,		/* open */
	flawd_release		/* release */
};


/*
 * And now the modules code and kernel interface.
 */

/*
 * User-settable array of minor device sizes (in units of 1k)
 */
static int size[NR_FLAWD_MINOR] = {1024};
MODULE_PARM (size, "i");

int init_module (void) {
	int	i, len, sectors, bitmap_len;
	struct	flawd_device *flawd;
	char	*buffer, *bitmap;

	if (register_blkdev(MAJOR_NR, "flawd", &flawd_fops)) {
		printk("Unable to get major number %d for flawd device\n",
		       MAJOR_NR);
		return -EIO;
	}

	/*
	 * Set up the per-major-device tables
	 */
	blk_dev[MAJOR_NR].request_fn = DEVICE_REQUEST;
	memset(&flawd_sizes, 0, sizeof(flawd_sizes));
	memset(&flawd_blksizes, 0, sizeof(flawd_blksizes));
	blk_size[MAJOR_NR] = flawd_sizes;
	blksize_size[MAJOR_NR] = flawd_blksizes;

	/* 
	 * Initialise each requested flawd minor device
	 */
	for (i=0; i < NR_FLAWD_MINOR; i++) {
		flawd = flawd_minor_table + i;
		
		memset(flawd, 0, sizeof(struct flawd_device));
		flawd->devnr = MKDEV (MAJOR_NR, i);
		
		if (!size[i])
			continue;
		
		len = size[i] * 1024;
		sectors = size[i] * 2;
		bitmap_len = (sectors + 7) / 8;
		
		/*
		 * Allocate the ramdisk
		 */
		buffer = vmalloc (len);

		/*
		 * Also allocate a bad block bitmap, with one bit per
		 * sector
		 */
		if (buffer)			
			bitmap = vmalloc (bitmap_len);
		
		/*
		 * If both allocations were successful, set up the minor
		 * device
		 */
		if (buffer && bitmap) {
			memset (bitmap, 0, bitmap_len);
			flawd->nr_sectors = sectors;
			flawd->nr_bytes   = len;
			flawd->data	  = buffer;
			flawd->bad_block_map = bitmap;
			flawd_sizes[i]    = size[i];
		} else {
			/*
			 * If any allocations failed, do cleanup
			 */
			if (buffer)
				vfree (buffer);
		}
	}

	return 0;
}

void cleanup_module( void ) {
	int i;
	struct flawd_device *flawd;
	
	for (i=0; i < NR_FLAWD_MINOR; i++) {
		flawd = flawd_minor_table + i;
		if (!flawd->nr_bytes)
			continue;
		
		vfree (flawd->data);
		vfree (flawd->bad_block_map);
	}
	
	if (unregister_blkdev(MAJOR_NR, "flawd") != 0)
		printk("flawd: cleanup_module failed\n");
}
